"""Module providing the BookingSystem interface for Stratocore's PPMS."""

from logging import INFO

from datetime import datetime, date, time

from django.conf import settings
from django.utils.timezone import make_aware, make_naive

from loguru import logger as log
from logging_interceptor import setup_loguru_interceptor
from pyppms import ppms

from ..models import User, PpmsBooking, Reservation, DeliveryGroup, Issue


class BookingSystem:
    """Fetch and process information from PPMS and update the DIOP database.

    Attributes
    ----------
    interceptor_level : int
        The logging level to pass to the `setup_loguru_interceptor` call.
    cf_ref : int
        The PPMS core facility reference.
    """

    def __init__(self, interceptor_level=INFO):
        self.interceptor_level = interceptor_level
        log.debug("Instantiating a BookingSystem for 'PPMS'...")
        self.cf_ref = settings.PYPPMS_CORE_REF

        self._conn = None
        self._supersystems = None
        self._supersystems_ids = []

    @property
    def conn(self):
        """Return the PPMS connection, or create it if not yet connected.

        Returns
        -------
        pyppms.ppms.PpmsConnection
        """
        if self._conn is not None:
            return self._conn

        setup_loguru_interceptor(level=self.interceptor_level)
        cache_path = settings.PYPPMS_CACHE
        log.debug(f"Using [{cache_path}] as PyPPMS cache path.")
        api_url = settings.PYPPMS_URI
        api_key = settings.PYPPMS_API_KEY
        self._conn = ppms.PpmsConnection(
            api_url, api_key, cache=cache_path, cache_users_only=True
        )
        log.debug(f"Connected to PPMS at [{api_url}].")
        return self._conn

    @property
    def supersystems(self):
        """A dict mapping names to IDs of "supersystems" from PPMS.

        A "supersystem" is a virtual / meta group of systems comprised of two or
        more actual PPMS systems. Currently PPMS doesn't have a concept of
        booking just *any* slot in no-matter-what-exact-system, but we need this
        to reflect the random assignment of machines (i.e. there is no fixed
        mapping from a PPMS-system to a DIOP-machine).

        The method will query PPMS for all systems matching the "room" /
        "loca(lisa)tion" value defined in the `PumapiSystemsLocation`
        configuration setting and group them by the values specified in
        `PumapiSystemGroups`. That setting may contain multiple group names
        separated by a comma, each of them will result in a separate entry in
        the returned dict (unless empty).

        Returns
        -------
        dict
            A dict with the group name (the "supersystem" label) as the key and
            a list with the corresponding PPMS system IDs as value. Supersystems
            that do not have any related systems in PPMS will be skipped.

        Example
        -------
        >>> ppms = BookingSystem()
        >>> ppms.supersystems
        {
            "WestWingBookableSuperSystem": [37, 38, 39, 44],
            "HighlandsMetaSystem": [73, 74, 77, 78],
        }
        """
        if self._supersystems is not None:
            return self._supersystems

        self._supersystems = {}
        location_filter = settings.PYPPMS_LOC_FILTER
        name_groups = settings.PYPPMS_NAME_GROUPS
        log.trace(f"Using 'Room=[{location_filter}]' to identify systems in PPMS...")

        for name_group in name_groups:
            group_ids = self.conn.get_systems_matching(
                localisation=location_filter, name_contains=[name_group]
            )
            if not group_ids:
                log.warning(f"No systems matching name [{name_group}] found!")
                continue

            log.trace(f"System IDs matching name [{name_group}]: {group_ids}")
            self._supersystems[name_group] = group_ids

        return self._supersystems

    @property
    def supersystems_ids(self):
        """A list of all PPMS system IDs belonging to any super-system."""
        if self._supersystems_ids:
            return self._supersystems_ids

        for ids in self.supersystems.values():
            self._supersystems_ids += ids

        return self._supersystems_ids

    def _merge_reservations_all(self, username: str, category: str):
        """Merge adjacent reservations and update corresponding bookings.

        Parameters
        ----------
        username : str
            The username to look for reservations / bookings.
        category : str
            The delivery group related to the reservations.

        Returns
        -------
        int
            The count of merged reservations.
        """
        merged = self._merge_reservations_all_rec(username, category)
        if merged > 0:
            log.success(f"🗜 🗓🗓✅ Merged (and removed) {merged} reservations.")
        else:
            log.debug("🗓 🔍 No adjacent reservations found, nothing merged.")

        return merged

    def _merge_reservations_all_rec(self, username: str, category: str):
        merged = 0
        all_res = Reservation.objects.filter(user=username, deliverygroup=category)
        if len(all_res) < 2:
            return 0

        log.debug(f"🗓🗓🔍 Checking {len(all_res)} reservations of {username}...")
        for res in all_res:
            try:
                adjacent = Reservation.objects.get(
                    user=username,
                    deliverygroup=category,
                    t_start=res.t_end,
                )
                merged += self._merge_reservations(res, adjacent)
                log.debug("🗜 🗓🗓 ➿ Merged reservations, recursion necessary...")
                merged += self._merge_reservations_all_rec(username, category)
                break
            except Reservation.DoesNotExist:
                log.debug(f"🗓 🔍 No adjacent reservation for ({res.id}) found.")

        return merged

    def _merge_reservations(self, res_1: Reservation, res_2: Reservation):
        """Merge two adjacent reservations.

        Check if the end time of the first given reservation is matching the
        start time of the second one and, if yes, merge the second reservation
        into the first one (the second one will be removed).

        Parameters
        ----------
        res_1 : Reservation
            The first reservation to be merged. In case the merge is happening,
            this reservation will be extended and kept in the database.
        res_2 : Reservation
            The second reservation to be merged. In case the merge is happening,
            this reservation will be discarded after extending the first one.

        Returns
        -------
        int
            1 in case the reservations got merged, 0 otherwise.
        """
        if res_1.t_end != res_2.t_start:
            log.warning(
                f"💣🔥 Not merging non-adjacent reservations: [{res_1}] vs. [{res_2}]"
            )
            return 0

        log.info(f"🗓 🧲🗓 Merging adjacent reservations: [{res_1}] + [{res_2}]")
        res_1.t_end = res_2.t_end
        res_1.save()
        for booking in PpmsBooking.objects.filter(reservation=res_2):
            booking.reservation = res_1
            booking.save()
        log.debug(f"🗓🚮 Deleting merged reservation ({res_2.id})...")
        res_2.delete()
        log.success(f"🗓 🧲🗓✅ Done merging adjacent reservations: {res_1}")

        return 1

    def _populate_name_mapping(self):
        """Fetch users from the DIOP-DB and populate the PyPPMS name mapping."""
        log.trace("Populating the pyppms name mapping dict...")
        users = User.objects.filter(enabled=True, accessgroup_member=True)
        for user in users:
            # use the standard `fullname` attribute as a fallback:
            key = user.fullname

            # in case the DB has the `ppms_fullname` filled, it has priority:
            if user.ppms_fullname:
                key = user.ppms_fullname
                log.trace(f"Using `ppms_fullname` attribute: {key}")

            self.conn.fullname_mapping[key] = user.username
        # log.trace(self.conn.fullname_mapping)

    def _process_booking_updates(self):
        """Process changes to bookings after pulling from PPMS."""
        self._process_deleted_bookings_all()
        self._process_new_bookings_all()
        new = PpmsBooking.objects.filter(sync_state="N").count()
        if new:
            msg = f"Still {new} bookings in state 'New' after processing!"
            log.warning(f"💫🎫💥 {msg}")
            issue = Issue(description=msg, severity=2)
            issue.save()

    def _split_reservation(self, booking):
        res = booking.reservation
        log.info(f"🗓🪓 Splitting reservation {res} on booking {booking}...")
        split_start = booking.t_start
        split_end = booking.t_end
        res_new = Reservation(
            user=res.user,
            deliverygroup=res.deliverygroup,
            t_start=split_end,
            t_end=res.t_end,
        )
        res_new.save()

        res.t_end = split_start
        res.save()
        log.success(f"✨🗓🗓 New reservations: {res} 🚧 {res_new}")

        updated = 0
        for linked in PpmsBooking.objects.filter(reservation=res):
            if linked.t_start < res_new.t_start:
                log.debug(f"Not updating, belongs to old res: {linked}")
                continue
            log.info(f"🎫🔗🗓  Updating link of [{linked}] to [{res_new.id}]...")
            linked.reservation = res_new
            linked.save()
            updated += 1
        log.info(f"🎫🔗🗓✅ Updated {updated} links to reservation [{res_new.id}].")

        log.success(
            f"🗓🪓✅ Done splitting reservation {res}, additional "
            f"reservation {res_new}"
        )

    def _process_deleted_bookings(self, system_ids, username):
        """Process bookings from a given user that are marked as `deleted`.

        First, all bookings having their `sync_state` field set to `deleted` and
        that are matching the given user and the list of PPMS system IDs will be
        fetched from the DB.

        Then they will be checked if they are linked to a reservation (meaning
        they belong to a user account known to the DIOP-DB *and* have been
        processed before).

        In case a booking is not linked to a reservation, it will be deleted
        from the bookings table (NOT from PPMS!).

        Otherwise, the booking will be checked if it covers the entire
        reservation (reservation will be deleted), if it matches the start or
        the end (reservation will be shortened), or if it's fully enclosed by
        the reservation (requiring the reservation to be split).

        Parameters
        ----------
        system_ids : list(str)
            A list of PPMS system IDs to fetch the bookings for.
        username : str
            The PPMS username to check bookings for.

        Returns
        -------
        int
            The number of deleted bookings that has been processed.
        """
        deleted = PpmsBooking.objects.filter(
            username=username, sync_state="D", system_id__in=system_ids
        )
        if len(deleted) == 0:
            log.debug("⏹🚮🎫 All deleted bookings have been processed!")
            return 0

        num = len(deleted)
        log.info(f"▶🚮🎫 Processing {num} DELETED bookings of 🙋 [{username}]...")
        booking = deleted[0]
        if booking.reservation is None:
            log.info(f"Booking 🎫 is unlinked, deleting it: {booking}")

        else:
            log.debug(f"🚮🎫 Processing deleted booking: {booking}")

            # TODO: review the entire "else" block to make sure we haven't forgotten
            # conditions / edge cases
            res = booking.reservation
            log.debug(f"Linked 🎫🔗🗓  reservation: {res}")
            if booking.t_start == res.t_start and booking.t_end == res.t_end:
                log.info("🗓🚮 Start AND end matching, DELETING reservation...")
                res.delete()
            elif booking.t_end == res.t_end:
                log.info("🗓⏮ End matching, SHORTENING reservation...")
                res.t_end = booking.t_start
                res.save()
            elif booking.t_start == res.t_start:
                log.info("🗓⏭ Start matching, SHIFTING reservation START...")
                res.t_start = booking.t_end
                res.save()
            else:
                self._split_reservation(booking)

        log.success(
            f"🚮🎫✅ Processed booking ({booking.id}), removing it from the DB..."
        )
        # FIXME: this will fail if the booking has been deleted through a FK
        # relation automatically (reservation has been deleted)!
        booking.delete()

        return self._process_deleted_bookings(system_ids, username) + 1

    def _process_deleted_bookings_all(self):
        """Process all bookings having `sync_state` set to `deleted`.

        The method iterates over all users having any booking where `sync_state`
        is set to `deleted` and calls the corresponding method to process those
        bookings for each PPMS supersystem group.

        Returns
        -------
        int
            The total count of processed deleted bookings.
        """
        log.trace("▶🚮🎫Processing 'deleted' bookings...")
        usernames = (
            PpmsBooking.objects.filter(sync_state="D")
            .values_list("username", flat=True)
            .distinct()
        )
        total = 0
        for category, system_ids in self.supersystems.items():
            log.debug(f"Category: {category}")
            cat_total = 0
            for username in usernames:
                num = self._process_deleted_bookings(system_ids, username)
                log.debug(f"⏹🚮🎫 Processed {num} DELETED bookings of {username}...")
                cat_total += num

            if cat_total == 0:
                continue

            total += cat_total
            log.info(
                f"🔄🚮🎫✅ Processed {cat_total} booking updates in state 'deleted' "
                f"for [{category}], IDs {system_ids}."
            )

        return total

    def _process_new_bookings(self, category, system_ids, username):
        """Process `new` bookings of a given user and PPMS supersystem.

        First, the function makes sure the given user doesn't have any bookings
        marked as `deleted`, otherwise processing will be stopped as those ones
        *must* be treated before.

        Then all PPMS-bookings of the user marked as `new` are retrieved which
        are matching the PPMS system IDs making up the "supersystem".

        Next, each of those bookings is compared against *all* reservations to
        decide if the new booking is extending that reservation (shifting the
        reservation's start or end time) or if it's actually overlapping
        (meaning a violation of the no-double-bookings rule).

        In case none of the previous conditions is matching, the booking is
        actually a truly new one that will be used to create a new reservation.

        Eventually (after processing all new bookings), the reservations are
        checked and merged if the updates resulted in adjacent reservations.

        Parameters
        ----------
        category : str
            The name of a category (=delivery group).
        system_ids : list(int)
            The PPMS system IDs making up this category ("supersystem").
        username : str
            The username owning the bookings.
        """
        user = User.objects.get(username=username)

        deleted = PpmsBooking.objects.filter(
            username=username, sync_state="D", system_id__in=system_ids
        ).count()
        if deleted > 0:
            desc = (
                f"User [{username}] is having {deleted} pending bookings in "
                "state 'DELETED', cannot process this user's new bookings!"
            )
            log.error(desc)
            issue = Issue(user=user, description=desc, severity=3)
            issue.save()
            return

        bookings_new = PpmsBooking.objects.filter(
            username=username, sync_state="N", system_id__in=system_ids
        )
        log.debug(f"💫🎫 Processing NEW bookings of 🙋 [{username}]...")

        end = start = double = new = 0
        for booking in bookings_new:
            is_new_reservation = True

            # compare new booking against all existing reservations:
            existing_reservations = Reservation.objects.filter(
                user=username, deliverygroup=category
            )
            for res in existing_reservations:
                # check if booking start matches the end of the reservation:
                if booking.t_start == res.t_end:
                    res.t_end = booking.t_end
                    res.save()
                    log.success(f"🗓⏭  New reservation end: {res}")
                    booking.reservation = res
                    booking.sync_state = "K"
                    booking.save()
                    is_new_reservation = False
                    end += 1
                    break

                # check if booking end matches the start of the reservation:
                if booking.t_end == res.t_start:
                    res.t_start = booking.t_start
                    res.save()
                    log.success(f"🗓⏮  New reservation start: {res}")
                    booking.reservation = res
                    booking.sync_state = "K"
                    booking.save()
                    is_new_reservation = False
                    start += 1
                    break

                # check if booking overlaps with reservation (double-booking):
                if booking.t_start in res.t_range or booking.t_end in res.t_range:
                    log.warning(
                        f"🎫💥 booking [{booking}] overlaps with 🗓  reservation "
                        f"[{res}], NEW booking will be DISCARDED!"
                    )
                    is_new_reservation = False
                    # FIXME: what other actions are needed on double bookings?
                    # - remove entire reservation (if possible)
                    # - add a issue record
                    # - send a user warning, or defer it to the notifications dispatcher
                    double += 1
                    break

            # bookings still not processed are actually NEW reservations:
            if is_new_reservation:
                deliverygroup = DeliveryGroup.objects.get(dg_name=category)
                new_res = Reservation(
                    user=user,
                    deliverygroup=deliverygroup,
                    t_start=booking.t_start,
                    t_end=booking.t_end,
                )
                new_res.save()
                log.success(f"✨🗓 New reservation: {new_res}")
                booking.reservation = new_res
                booking.sync_state = "K"
                booking.save()
                new += 1

        log.debug(f"💫🎫✅ Processed {len(bookings_new)} NEW bookings.")

        # now that all new bookings have been processed we need to check if the
        # new situation is having "adjacent" reservations that should be merged:
        merged = self._merge_reservations_all(username, category)

        return new, start, end, merged, double

    def _process_new_bookings_all(self):
        """Process all bookings having `sync_state` set to `new`."""
        usernames = (
            PpmsBooking.objects.filter(sync_state="N")
            .values_list("username", flat=True)
            .distinct()
        )

        for category, system_ids in self.supersystems.items():
            log.debug(f"Category: {category}, IDs: {system_ids}")
            for username in usernames:
                # TODO: use the returned values on processed bookings:
                self._process_new_bookings(category, system_ids, username)

    def _update_bookings(self, runningsheet, date):
        """Update the DB with PPMS bookings from the given runningsheet and date.

        Parameters
        ----------
        runningsheet : list(pyppms.booking.PpmsBooking)
            The runningsheet retrieved from PPMS.
        date : datetime.datetime
            The date associated to the runningsheet.

        Returns
        -------
        int, int
            A tuple with counts of success / failed updates.
        """
        # first mark all existing bookings for the given date as "deleted"
        PpmsBooking.mark_all_deleted(date)

        okay = fail = skipped = 0

        for booking in runningsheet:
            if booking.system_id not in self.supersystems_ids:
                log.trace(f"⭕🎫 Skipping, system ID not in list: {booking.desc}")
                skipped += 1
                continue

            log.debug(f"🔎🎫 Checking booking in the DB: {booking.desc}")

            d_booking, created = PpmsBooking.objects.get_or_create(
                username=booking.username,
                system_id=booking.system_id,
                t_start=make_aware(booking.starttime),
                t_end=make_aware(booking.endtime),
                # NOTE: we must specify *EXACTLY* those fields that are members
                # of the `UniqueConstraint` of the model
            )
            # now adjust 'sync_state' and possibly other fields that are not part of the
            # constraint:
            # `sync_state` of existing entries has been set to "D" (deleted) by the
            # `mark_all_deleted()` call above, for entries that have been created newly
            # here it will be set to the model's default value "N" (new)
            if not created:
                d_booking.sync_state = "K"
                log.debug("🔄🎫 Refreshed booking (sync_state=known) in the DB.")
            else:
                log.success(f"✨🎫 Created NEW booking in the DB: {d_booking}")
            d_booking.ppms_session = booking.session
            d_booking.save()
            okay += 1

        if skipped:
            log.debug(f"⭕🎫 Skipped {skipped} bookings (system IDs not in list).")

        return okay, fail

    def _update_ppms_user_details(self, username, ppms_name, ppms_group):
        """Update the PPMS details of a user record in the DIOP-DB.

        Parameters
        ----------
        username : str
            The username identifying the record in the DIOP-DB.
        ppms_name : str
            The "PPMS Full Name" to be set for the record.
        ppms_group : str
            The "PPMS Group" to be set for the record.

        Returns
        -------
        int, int
            A tuple with the number of updated and failed users.
        """
        log.debug(
            f"Updating 👩‍🔬👨‍🔬 user record for [{username}] with PPMS details: "
            f"[{ppms_name}]/[{ppms_group}]"
        )
        okay = fail = 0
        try:
            d_user = User.objects.get(username=username)
            if d_user.ppms_fullname != ppms_name or d_user.ppms_group != ppms_group:
                d_user.ppms_fullname = ppms_name
                d_user.ppms_group = ppms_group
                d_user.save()
            okay += 1
        except User.DoesNotExist:
            # this one should explicitly be re-raised, as it indicates some potentially
            # severe underlying problem:
            raise
        except Exception as err:  # pylint: disable-msg=broad-except
            # silence all other exceptions after issuing a warning message:
            log.warning(f"Unexpected exception: {err}")
            fail += 1

        return okay, fail

    def pull_bookings(self, dt=date.today()):
        """Fetch PPMS bookings, put them into the DIOP-DB and process them.

        After pulling all bookings for the given date (using the PPMS
        runningsheet) and putting them into the DIOP-DB, they will be processed
        and the DIOP reservations will be updated accordingly.

        Parameters
        ----------
        dt : datetime.date, optional
            The date for which to fetch bookings, by default `date.today.()`.

        Returns
        -------
        int, int
            A tuple with counts of success / failed booking updates. Please note
            that this doesn't contain any numbers on the changes to
            *reservations* occurring from processing the booking changes!

        Notes
        -----
        For being able to register a new Reservation in the database, the
        corresponding DeliveryGroup has to exist already!
        """
        log.debug(f"▶🔄📋🎫 Pulling PPMS bookings for {dt}...")
        self._populate_name_mapping()
        runningsheet = self.conn.get_running_sheet(
            core_facility_ref=self.cf_ref,
            date=datetime.combine(dt, time.min),
            ignore_uncached_users=True,
            localisation=settings.PYPPMS_LOC_FILTER,
        )
        log.debug(f"Got {len(runningsheet)} bookings for {dt}")
        okay, fail = self._update_bookings(runningsheet, dt)
        log.success(f"⏹🔄📋🎫✅ Fetched {okay} bookings for {dt}.")

        self._process_booking_updates()

        return okay, fail

    def pull_users(self, force_refresh=False):
        """Query PPMS on known users and update the DIOP users table.

        Parameters
        ----------
        force_refresh : bool, optional
            Passed on to `pyppms.ppms.PpmsConnection.get_user()`, by default False.

        Returns
        -------
        int, int
            A tuple with counts of success / failed updates.
        """
        d_users = User.objects.filter(accessgroup_member=True, enabled=True)
        log.debug(
            f"▶🔄👩‍🔬👨‍🔬 Fetching {len(d_users)} active user details from PPMS..."
        )
        ppms_users = []
        for user in d_users:
            try:
                ppms_user = self.conn.get_user(user.username, skip_cache=force_refresh)
                ppms_users.append(ppms_user)
            except KeyError:
                # potentially there are many users that do NOT have an account in the
                # booking system, so we're only trace-logging this:
                log.trace(f"👩‍🔬👨‍🔬 Skipping [{user.username}]: not existing in PPMS!")
        log.debug(f"🔄👩‍🔬👨‍🔬 Got {len(ppms_users)} user details from PPMS.")
        okay = fail = 0
        for ppms_user in ppms_users:
            try:
                u_okay, u_fail = self._update_ppms_user_details(
                    username=ppms_user.username,
                    ppms_name=ppms_user.fullname,
                    ppms_group=ppms_user.ppms_group,
                )
                okay += u_okay
                fail += u_fail
            except Exception as err:  # pylint: disable-msg=broad-except
                log.warning(
                    f"Failed updating PPMS details for [{ppms_user.username}]: {err}"
                )
                fail += 1

        log.debug(f"⏹🔄👩‍🔬👨‍🔬 Finished updating users: okay={okay}, failed={fail}.")
        return okay, fail



# Generated by Django 4.1.7 on 2023-04-14 13:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0003_alter_siteconfig_options_alter_machine_fqdn_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="DeliveryGroup",
            fields=[
                (
                    "dg_name",
                    models.CharField(
                        max_length=32,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Name of the Delivery Group (DG)",
                    ),
                ),
                (
                    "host_prefix",
                    models.CharField(
                        max_length=32, verbose_name="Hostname prefix identifying the DG"
                    ),
                ),
                (
                    "booking_required",
                    models.BooleanField(
                        default=False, verbose_name="DG requires booking"
                    ),
                ),
            ],
            options={
                "verbose_name": "Delivery Group",
            },
        ),
    ]



from django.urls import path

from . import views


urlpatterns = [
    path("", views.dg_states, name="index"),
    path("machines/basic/", views.MachineListBasic.as_view(), name="machines_basic"),
    path("machines/", views.MachineList.as_view(), name="machine_list"),
    path("sessions/", views.SessionList.as_view(), name="session_list"),
    path("session/<int:uid>/", views.SessionDetails.as_view(), name="session_details"),
    path("update/<str:item>/", views.update_item, name="update_item"),
    path(
        "issues_unresolved/",
        views.status_issues_unresolved,
        name="status_issues_unresolved",
    ),
    path(
        "last_update/<str:item>/",
        views.last_update_item,
        name="last_update_item",
    ),
    path("ops/", views.OperatorsDash.as_view(), name="operators_dash"),
    path(
        "reservation/<int:id>/",
        views.ReservationDetails.as_view(),
        name="reservation_details",
    ),
]



"""Diop PpmsBooking model."""

from datetime import timedelta, datetime, time

from loguru import logger as log

from django.db import models
from django.utils.timezone import make_aware

from ..utils import naive_minutes
from .reservation import Reservation


class PpmsBooking(models.Model):
    class Meta:
        verbose_name = "PPMS Booking"
        constraints = [
            models.UniqueConstraint(
                fields=["username", "system_id", "t_start", "t_end"],
                name="unique_booking",
            )
        ]

    class SyncState(models.TextChoices):
        DELETED = "D"
        NEW = "N"
        KNOWN = "K"

    # NOTE: username is NOT a FK - it stores the data just as returned by PPMS
    username = models.CharField(max_length=32, verbose_name="PPMS Username")
    system_id = models.IntegerField("PPMS System ID")
    t_start = models.DateTimeField("Booking Start", null=False, blank=False)
    t_end = models.DateTimeField("Booking End", null=False, blank=False)
    ppms_session = models.CharField(
        "PPMS Session ID", max_length=8, null=True, blank=True
    )
    sync_state = models.CharField(
        max_length=1,
        choices=SyncState.choices,
        default=SyncState.NEW,
    )
    reservation = models.ForeignKey(
        Reservation,
        on_delete=models.CASCADE,
        help_text="Reservation to which this booking belongs.",
        null=True,
        blank=True,
    )

    def __str__(self):
        return (
            f"({self.pk}) {self.username}@{self.system_id} "
            f"[{naive_minutes(self.t_start)} -- {naive_minutes(self.t_end)}]"
        )

    @classmethod
    def mark_all_deleted(cls, date):
        """Set `sync_state` to `D` for all bookings on the given day.

        Parameters
        ----------
        date : datetime.datetime
            The date for which the `sync_state` should be set to "deleted".
        """
        date = datetime.combine(date, time.min)
        t0 = make_aware(date)
        t1 = make_aware(date + timedelta(days=1))
        log.debug(f"Setting bookings to DELETED from [{t0}] to [{t1}].")
        cls.objects.filter(t_start__gte=t0, t_end__lte=t1).update(sync_state="D")



"""Diop MessageTemplate model."""

import re

from box import Box

from django.conf import settings
from django.db import models
from django.template import Template, Context

from .. import site


class MessageTemplate(models.Model):
    class Meta:
        verbose_name = "Message Template"

    class MessageCategory(models.TextChoices):
        QUESTION = "Q"
        INFO = "I"
        WARNING = "W"
        ERROR = "E"

    name = models.CharField(
        "Template Name",
        max_length=64,
        primary_key=True,
    )
    category = models.CharField(
        max_length=1,
        choices=MessageCategory.choices,
        default=MessageCategory.INFO,
        verbose_name="Message Category",
    )
    subject = models.CharField(
        max_length=256,
        null=False,
        blank=False,
        help_text="The template subject / title.",
    )
    body = models.TextField(
        null=False,
        blank=False,
        help_text="The template body / content.",
    )

    def __str__(self):
        return self.name

    @classmethod
    def render_message(cls, template: str, context: dict, prefix: bool = True):
        """Helper to render a message from a DB template.

        Parameters
        ----------
        template : str
            The name of the template to use for the message.
        context : dict
            The context dict required to render the template.
        prefix : bool, optional
            If set to `True` a site-specific label will be prefixed to the subject,
            enclosed in square brackets, otherwise the subject will not be touched
            (by default True).

        Returns
        -------
        Box
            A box object containing the strings for the message category, subject,
            body, subject_raw, body_raw with the latter two having all Django
            template filters (like `naturaltime`) stripped from the template before
            rendering them. This is useful to produce identical messages independent
            of external circumstances (like the current time and so on) that can be
            used for determining if a message has been generated already before. The
            attributes of the box are:
            - `subject`
            - `body`
            - `subject_raw`
            - `body_raw`
            - `category`
        """
        if not context.get("support_contact"):
            context["support_contact"] = settings.DIOP_SUPPORT_CONTACT

        if not context.get("site_name"):
            context["site_name"] = site.name()

        if not context.get("booking_uri"):
            context["booking_uri"] = site.booking_uri()

        msg_template = cls.objects.get(pk=template)

        subj_template = msg_template.subject
        body_template = msg_template.body

        if prefix:  # prefix the subject with the site-name in square brackets
            subj_template = f"[{site.name()}] {subj_template}"

        subject = Template(subj_template).render(Context(context))
        body = Template(body_template).render(Context(context))

        pattern = re.compile(r"\| *[a-z]* *}}")
        subj_template_raw = re.sub(pattern, "}}", subj_template)
        body_template_raw = re.sub(pattern, "}}", body_template)
        subject_raw = Template(subj_template_raw).render(Context(context))
        body_raw = Template(body_template_raw).render(Context(context))

        return Box(
            {
                "subject": subject,
                "body": body,
                "subject_raw": subject_raw,
                "body_raw": body_raw,
                "category": msg_template.category,
            }
        )



"""Diop Issue model."""

from django.db import models
from django.utils.timezone import make_naive

from loguru import logger as log

from .task import Task
from .user import User


class Issue(models.Model):
    class Level(models.IntegerChoices):
        LOW = 0, "Low"
        MEDIUM = 1, "Medium"
        HIGH = 2, "High"
        CRITICAL = 3, "Critical"
        MELTDOWN = 4, "Meltdown"

    severity = models.SmallIntegerField(
        default=0,
        choices=Level.choices,
        help_text="Values from 0 to 4, higher means more critical.",
    )
    timestamp = models.DateTimeField(
        auto_now_add=True,
        help_text="Time the issue occurred.",
    )
    resolved = models.BooleanField(
        default=False,
        help_text="Flag indicating the issue is resolved.",
    )
    task = models.ForeignKey(
        Task,
        on_delete=models.CASCADE,
        help_text="Related task causing the issue (if applicable).",
        null=True,
        blank=True,
    )
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        help_text="Related user (if applicable).",
        null=True,
        blank=True,
    )
    description = models.TextField()

    def __str__(self):
        severity = self.get_severity_display()
        timestamp = self.timestamp
        try:
            timestamp = make_naive(self.timestamp).strftime("%H:%M:%S")
        except Exception as err:
            log.error(f"Failed to format timestamp [{timestamp}]: {err}")
        user = f", user=[{self.user.username}]" if self.user else ""
        task = f", task=[{self.task.short}]" if self.task else ""

        return f"#{self.id} [{severity}] ({timestamp}){user}{task}"



"""URL Configuration for `diop_core`.

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/4.1/topics/http/urls/

Examples
--------
Function views
1. Add an import:  `from my_app import views`
2. Add a URL to urlpatterns:  `path('', views.home, name='home')`

Class-based views
1. Add an import:  `from other_app.views import Home`
2. Add a URL to urlpatterns:  `path('', Home.as_view(), name='home')`

Including another URLconf
1. Import the include() function:  `from django.urls import include, path`
2. Add a URL to urlpatterns:  `path('blog/', include('blog.urls'))`
"""

from django.contrib import admin
from django.contrib.auth.views import LoginView
from django.urls import include, path

urlpatterns = [
    path("accounts/", include("django.contrib.auth.urls")),
    path("admin/", admin.site.urls),  # Django's built-in admin module
    path("", include("diop.urls")),
]



"""Diop DeliveryGroup model."""

from loguru import logger as log

from django.db import models

from .user import User


class DeliveryGroup(models.Model):
    dg_name = models.CharField(
        "Group Name",
        max_length=32,
        primary_key=True,
        help_text="Name of the Delivery Group (DG).",
    )
    host_prefix = models.CharField(
        "Hostname Prefix",
        max_length=32,
        help_text=(
            "Hostnames starting with this prefix will be considered "
            "members of the DG."
        ),
        default="",
        blank=True,
    )
    booking_required = models.BooleanField(
        "Booking Required",
        default=False,
        help_text="Sessions in this DG require a valid booking.",
    )
    access_users = models.ManyToManyField(
        to=User,
        blank=True,
        help_text="Users that are explicitly granted access to this DG.",
    )
    unique_session = models.BooleanField(
        default=True,
        help_text=(
            "Allow only one session per user across *ALL* groups having this "
            "flag enabled (enforced via a housekeeping task)."
        ),
    )

    class Meta:
        verbose_name = "Delivery Group"

    def __str__(self):
        return self.dg_name

    @classmethod
    def get_or_new(cls, dg_name: str):
        """Get a DG from the DB or create a new one if it doesn't exist.

        Parameters
        ----------
        dg_name : str
            The name of the delivery group.

        Returns
        -------
        diop.models.DeliveryGroup, bool
            A tuple consisting of the delivery group DB record and a boolean flag
            indicating if the record has been newly created.
        """
        log.debug(f"💫 Checking for delivery group: {dg_name}")
        d_dg, created = cls.objects.get_or_create(dg_name=dg_name)
        if created:
            log.warning(f"💫 Created NEW delivery group: [{d_dg}] ✨")
        else:
            log.debug(f"💫 Using existing DG [{d_dg}] 🆗")
        return d_dg, created



# Generated by Django 4.2.2 on 2023-07-11 21:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0029_alter_deliverygroup_access_users_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="MessageTemplate",
            fields=[
                (
                    "name",
                    models.CharField(
                        max_length=64,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Template Name",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        choices=[
                            ("Q", "Question"),
                            ("I", "Info"),
                            ("W", "Warning"),
                            ("E", "Error"),
                        ],
                        default="I",
                        max_length=1,
                        verbose_name="Message Category",
                    ),
                ),
                (
                    "subject",
                    models.CharField(
                        help_text="The template subject / title.", max_length=256
                    ),
                ),
                ("body", models.TextField(help_text="The template body / content.")),
            ],
            options={
                "verbose_name": "Message Template",
            },
        ),
    ]



"""Diop SiteLog model."""

from django.db import models

from .user import User
from .deliverygroup import DeliveryGroup
from .machine import Machine
from .session import Session
from .task import Task
from .notification import Notification
from .reservation import Reservation


class SiteLog(models.Model):
    """Log entries of status changes, action requests and such.

    Entries in this table will reflect the site's history with status changes of
    sessions and machines, notifications, force-disconnects, force-terminates
    and more.
    """

    class Meta:
        verbose_name = "Site Log"

    name = models.CharField("Name", max_length=128)
    parameters = models.CharField("Parameters", max_length=512)
    time = models.DateTimeField(
        auto_now_add=True,
        help_text="Time the action was triggered.",
    )
    details = models.TextField()
    user = models.ForeignKey(User, on_delete=models.SET_NULL, null=True, blank=True)
    deliverygroup = models.ForeignKey(
        DeliveryGroup, on_delete=models.SET_NULL, null=True, blank=True
    )
    machine = models.ForeignKey(
        Machine, on_delete=models.SET_NULL, null=True, blank=True
    )
    session = models.ForeignKey(
        Session, on_delete=models.SET_NULL, null=True, blank=True
    )
    task = models.ForeignKey(Task, on_delete=models.SET_NULL, null=True, blank=True)
    notification = models.ForeignKey(
        Notification, on_delete=models.SET_NULL, null=True, blank=True
    )
    reservation = models.ForeignKey(
        Reservation, on_delete=models.SET_NULL, null=True, blank=True
    )



"""Adapter module to work with PSyTricks data."""

from os import environ

from loguru import logger as log
from psytricks.wrapper import ResTricksWrapper

from django.conf import settings
from django.utils.functional import LazyObject

from .. import site
from ..models import Machine, Session, SessionProperties, SiteLog, User, DeliveryGroup


class ResTricksClient(LazyObject):
    """LazyObject wrapper for the ResTricksWrapper class.

    By encapsulating the psytricks.wrapper.ResTricksWrapper in a LazyObject,
    instantiation of the object is deferred until its first actual usage.
    """

    def _setup(self):
        if "PDOC_DISPLAY_ENV_VARS" in environ.keys():
            # do NOT instantiate the wrapper if the code is only being processed
            # by pdoc (using the `PDOC_DISPLAY_ENV_VARS` environment variable
            # as an indicator)
            return

        log.debug("Instantiating ResTricksClient...")
        self._wrapped = ResTricksWrapper(
            base_url=settings.PSYTRICKS_BASE_URL,
            verify=settings.PSYTRICKS_VERIFY,
            lazy=True,
        )
        if settings.DIOP_DRY_RUN:
            self._wrapped.read_only = True


broker = ResTricksClient()


def process_session_info(psy_session: dict):
    """Update the DB on sessions and its properties from a session details dict.

    Parameters
    ----------
    psy_session : dict
        A dict with session details as returned by
        `psytricks.wrapper.ResTricksWrapper.get_sessions()`.

    Returns
    -------
    bool, bool
        A tuple of boolean values indicating if a new session record (1st
        element) or a new session properties record (2nd element) has been
        created from the session details.
    """
    # log.warning(psy_session)

    # FIXME: sometimes Citrix returns "partial" sessions for currently unknown
    # reasons, which won't contain all the details - resulting in an exception
    # being raised. Rather have a proper session object in the psytricks package
    # (e.g. that has an attribute `.valid` or similar) and use that one here
    # than blindly accessing dict fields that may not exist!

    d_session, created_s = Session.is_up_to_date(psy_session)

    defaults = {
        "client_address": psy_session["ClientAddress"],
        "client_name": psy_session["ClientName"],
        "client_version": psy_session["ClientVersion"],
    }

    d_session_props, created_p = SessionProperties.get_or_new(
        session=d_session,
        t_change=psy_session["SessionStateChangeTime"],
        state=psy_session["SessionState"],
        defaults=defaults,
    )

    log.debug(f"🎬✅ DONE updating details on [{d_session}]")
    return created_s, created_p


def process_machine_info(psy_machine: dict):
    # FIXME: see the note about incomplete details in `process_session_info` and
    # go for a similar strategy here!

    machine = psy_machine
    d_machine, _ = Machine.get_or_new(
        fqdn=machine["DNSName"],
        dg_name=machine["DesktopGroupName"],
        state=machine["SummaryState"],
    )

    # in case none of the machine's attributes changed we use .update() with
    # filtering for the PK to avoid the post-save signal from being triggered
    # exclusively for setting that field (otherwise each call to this function
    # would also trigger that signal):
    if (
        d_machine.state == machine["SummaryState"]
        and d_machine.powerstate == machine["PowerState"]
        and d_machine.maintenance == machine["InMaintenanceMode"]
        and d_machine.registration == machine["RegistrationState"]
        and d_machine.agent == machine["AgentVersion"]
    ):
        Machine.objects.filter(fqdn=machine["DNSName"]).update(updated=True)
        log.debug(f"🖥 ✅ Machine status unchanged: [{d_machine}]")
        return

    d_machine.state = machine["SummaryState"]
    d_machine.powerstate = machine["PowerState"]
    d_machine.maintenance = machine["InMaintenanceMode"]
    d_machine.registration = machine["RegistrationState"]
    d_machine.agent = machine["AgentVersion"]
    d_machine.updated = True
    d_machine.save()
    log.info(f"🖥 ✅ DONE updating details on [{d_machine}]")


def set_maintenance(disable_on, enable_on):
    """Enable or disable maintenance mode for a given machine.

    Parameters
    ----------
    disable_on : list(diop.models.machine.Machine)
        A list (or other iterable) of machine objects for which the maintenance
        mode should be DISABLED.
    enable_on : list(diop.models.machine.Machine)
        A list (or other iterable) of machine objects for which the maintenance
        mode should be ENABLED.

    Notes
    -----
    Maintenance mode on a machine is NOT enforced if the DIOP DB status suggests
    the machine is already in the requested mode! If the DB is outdated this
    will lead to incorrect actions.
    """
    for d_machine in enable_on:
        if d_machine.maintenance:
            log.trace(f"Not changing maintenance, already set to [True]: {d_machine}")
            continue

        if site.in_dry_run_mode(f"ENABLING maintenance mode for: {d_machine}"):
            continue

        log.info(f"ENABLING maintenance mode for: {d_machine}")
        broker.set_maintenance(machine=d_machine.fqdn, disable=False)

    for d_machine in disable_on:
        if not d_machine.maintenance:
            log.trace(f"Not changing maintenance, already set to [False]: {d_machine}")
            continue

        if site.in_dry_run_mode(f"DISABLING maintenance mode for: {d_machine}"):
            continue

        log.info(f"DISABLING maintenance mode for: {d_machine}")
        broker.set_maintenance(machine=d_machine.fqdn, disable=True)


def msg_category_to_psy_style(category: str):
    """Map a message category to a PSyTricks `style` attribute."""
    mapping = {
        "Q": "Question",
        "I": "Information",
        "W": "Exclamation",
        "E": "Critical",
    }
    return mapping[category]


def send_popup(machine: str, subject: str, body: str, category: str):
    """Deliver a pop-up message to a machine.

    Can be passed as a callable to
    `diop.models.notification.Notification.new_popup()` and similar.

    Parameters
    ----------
    machine : str
        The FQDN of the machine to send the message to.
    subject : str
        The message subject / title.
    body : str
        The message body / content.
    category : str
        The MessageTemplate category, will be mapped to a PSyTricks `style`.

    Returns
    -------
    bool
        True in case the pop-up was sent, False otherwise.
    """
    log.debug(f"🖥🎬📢 <send_message> ⏩ [{subject}]@[{machine}]")
    try:
        # FIXME: make dry-run aware!
        broker.send_message(
            machine=machine,
            message=body,
            title=subject,
            style=msg_category_to_psy_style(category),
        )
    except Exception as err:
        log.warning(f"Sending pop-up message failed: {err}")
        return False

    return True


def update_diop_dg_access(dg_name: str):
    """Fetch DG access users via PSyTricks and update the Diop DB accordingly.

    Parameters
    ----------
    dg_name : str
        The name of the delivery group to update.
    """
    dg = DeliveryGroup.objects.get(dg_name=dg_name)

    psy_users = get_dg_access(dg_name=dg_name, strip_domain=True)

    d_users = []
    for psy_user in psy_users:
        try:
            d_users.append(User.objects.get(username=psy_user))
        except:  # ruff: noqa: E722 (bare-except)
            log.info(f"Can't find user [{psy_user}], skipping!")

    dg.access_users.set(d_users)


def get_dg_access(dg_name: str, strip_domain: bool = False) -> list:
    r"""Get unique usernames having access to the given delivery group.

    Parameters
    ----------
    dg_name : str
        The name of the delivery group.
    strip_domain : bool, optional
        If set to `True` the returned list will contain "account names" only,
        meaning the `DOMAINNAME\\` prefix will be stripped.

    Returns
    -------
    list
    """
    usernames = []
    current = broker.get_access_users(dg_name)

    if not current:
        return []

    for user in current:
        name = str(user["Name"])
        if strip_domain:
            try:
                name = name.split("\\")[1]
            except:  # ruff: noqa
                pass
        if name not in usernames:
            usernames.append(name)

    usernames.sort()

    return usernames


def set_dg_access(dg_name: str, usernames: list, keep_common: bool = False):
    """Set access to a DG to be exactly the given list of usernames.

    Check the users currently having access to the given delivery group and add
    / remove permissions so exactly the ones provided in the usernames list are
    permitted to have access (see the `keep_common` parameter description for
    the only exception).

    If the `DIOP_DRY_RUN` setting is active no actual changes to the DG access
    permissions will be issued!

    Parameters
    ----------
    dg_name : str
        The name of the delivery group.
    usernames : list
        The list of usernames that should be permitted to access the DG.
    keep_common : bool, optional
        If set to `True` the user (or group, for that matter) defined in the
        `DIOP_USER_AD_GROUP` setting will be added to the list of accounts that
        are allowed to access the DG. Mostly useful for testing.
    """
    dom_pfx = settings.DIOP_NTDOMAIN
    current = set(get_dg_access(dg_name))
    desired = set()
    for username in usernames:
        if username.startswith(dom_pfx):
            desired.add(username)
            continue
        desired.add(f"{dom_pfx}\\{username}")

    if keep_common:
        desired.add(f"{dom_pfx}\\{settings.DIOP_USER_AD_GROUP.lower()}")
        desired = set(desired)

    if desired == current:
        log.debug("Not changing DG access, given list matches current permissions.")
        return

    to_remove = current.difference(desired)

    log.info(f"Current accounts enabled for DG [{dg_name}]: {current}")
    log.info(f"Accounts to be enabled for DG [{dg_name}]: {desired}")
    log.info(f"Accounts to be DISABLED for DG [{dg_name}]: {to_remove}")

    if site.in_dry_run_mode("adjusting of DG access permissions"):
        return

    broker.set_access_users(group=dg_name, users=",".join(desired), disable=False)
    if to_remove:
        broker.set_access_users(group=dg_name, users=",".join(to_remove), disable=True)

    log.success(
        "Accounts now allowed for DG "
        f"[{dg_name}]: {get_dg_access(dg_name, strip_domain=True)}"
    )


def terminate_session(d_session: Session = None, psy_session_uid: int = None):
    """Request a Diop session to be terminated (machine will be rebooted).

    One of the two optional parameters [d_session, psy_session_uid] is REQUIRED
    to identify the session to be terminated - if both are None a RuntimeError
    will be raised.

    Parameters
    ----------
    d_session : Session, optional
        A Diop session object whose machine should be requested to restart.
    psy_session_uid : int, optional
        A Citrix session UID whose machine should be requested to restart.

    Returns
    -------
    dict
        The parsed `JSON` response of the request's response as delivered by the
        ResTricksWrapper.

    Raises
    ------
    RuntimeError
        Raised in case both optional parameters are None or in case a session
        with the given UID can't be found in the database.
    """
    if not d_session and not psy_session_uid:
        raise RuntimeError("One of [d_session, psy_session_uid] is required!")

    uid = psy_session_uid

    if not d_session:
        # raises a diop.models.session.Session.DoesNotExist if not existing:
        d_session = Session.objects.get(pk=uid)
        log.debug(f"Session [{uid}] is hosted on machine [{d_session.machine}].")

    res = restart_machine(machine=d_session.machine.fqdn)
    record = SiteLog(
        name="Restart requested",
        session=d_session,
        machine=d_session.machine,
        user=d_session.user,
        deliverygroup=d_session.machine.deliverygroup,
        details=res,
    )
    record.save()

    return res


def restart_machine(machine: str):
    """Request a restart (reboot) of the given machine.

    Parameters
    ----------
    machine : str
        The FQDN of the machine to be restarted.

    Returns
    -------
    dict
        The parsed `JSON` response of the request's response as delivered by the
        ResTricksWrapper.
    """
    # FIXME: make dry-run aware!
    return broker.perform_poweraction(machine=machine, action="restart")



"""DIOP tasks."""

from datetime import datetime, timedelta, date

from django.db.models import Count
from django.db.models.query import QuerySet
from django.conf import settings
from django.utils.timezone import make_aware, now

from loguru import logger as log

from .adapters.activedirectory import ActiveDirectory
from .adapters.ppms import BookingSystem
from .adapters import psytricks
from .decorators import fire_and_dont_forget, status_update_task
from .exceptions import UpdateMachinesError, UpdateSessionsError
from .models import (
    Machine,
    Session,
    Notification,
    User,
    SiteLog,
    DeliveryGroup,
    Reservation,
)
from .site import in_dry_run_mode
from .utils import send_support_email


_BROKER = psytricks.ResTricksClient()


@status_update_task
def update_machine_status():
    """Fetch machine states from Citrix and update DB."""
    db_model = Machine
    desc = "machine"
    _prepare_update(db_model, desc)
    okay, fail = _update_status(
        desc=desc,
        new_state=_BROKER.get_machine_status(),
        process_cb=psytricks.process_machine_info,
        details="DNSName",
        err_obj=UpdateMachinesError,
    )
    # TODO: check if all machines have been updated, otherwise add an entry in
    # the issues table
    log.success(f"🔁🖥🖥🖥  DONE updating machines (ok: {okay}, fail: {fail})")
    return okay, fail


@status_update_task
def update_session_status():
    """Fetch session states from Citrix and update DB."""
    db_model = Session
    desc = "session"
    _prepare_update(db_model, desc)
    okay, fail = _update_status(
        desc=desc,
        new_state=_BROKER.get_sessions(),
        process_cb=psytricks.process_session_info,
        details="Uid",
        err_obj=UpdateSessionsError,
    )
    _process_ended_sessions()
    log.success(f"🔁🎬🎬🎬 DONE updating sessions (ok: {okay}, fail: {fail})")
    return okay, fail


def _prepare_update(db_model, desc):
    """Prepare all entries in a given DB model for updating.

    Shortcut function to set the `updated` field to `False` for all entries in
    a given model / table.

    Parameters
    ----------
    db_model : django.db.models.Model
        The model to use
    desc : str
        A description to use in related messages.
    """
    count = db_model.objects.all().update(updated=False)
    log.debug(f"🖥🎬🖥🎬  Prepared {count} {desc} objects for updating...")


def _update_status(desc, new_state, process_cb, details, err_obj):
    """Update DB with current infrastructure status.

    Parameters
    ----------
    desc : str
        The description of what is being updated.
    new_state : list(dict)
        The parsed JSON containing the information used to updated the DB.
    process_cb : function
        The callback function to use for processing each item in `new_state`.
    details : str
        The key to use for extracting detail information on the item to be
        updated.
    err_obj : Exception
        The exception to raise in case any of the items fails updating.

    Returns
    -------
    okay, fail: int, int
        A tuple denoting how many entries were successfully updated and how
        many failed.

    Raises
    ------
    err_obj
        Raised in case one or more entries failed updating, exception type is
        as specified in the call.
    """
    log.debug(f"🖥🎬🖥🎬 Got details on {len(new_state)} {desc}s.")
    okay = fail = 0
    fail_details = []

    for entry in new_state:
        try:
            process_cb(entry)
            okay += 1
        except Exception as err:
            fail_details.append(entry[details])
            log.error(f"🖥  Updating {desc} [{entry[details]}] FAILED: {err}")
            fail += 1

    if fail > 0:
        raise err_obj(
            f"🖥  Updating DB status failed for one or more {desc}s!",
            fail_details,
            okay,
            fail,
        )

    log.info(f"🖥🎬🖥🎬  Updated details on {okay} {desc}s.")
    return okay, 0


def _process_ended_sessions():
    """Set session entries end timestamp for orphaned sessions."""
    sessions = Session.objects.filter(t_end=None).filter(updated=False)
    if not sessions:
        log.debug("🛑🎬🎬🎬 Checked for ended sessions, all good.")
        return

    for session in sessions:
        session.t_end = make_aware(datetime.now())
        session.updated = True
        session.save()
        log.success(f"🛑🎬 Registered {session} as ended in the DB!")
    log.info(f"🛑🎬🎬🎬✅ Updated {len(sessions)} ended session entries!")


@fire_and_dont_forget
def terminate_session(uid):
    """Terminate the session with the given UID.

    Parameters
    ----------
    uid : int
        A Citrix session UID whose machine should be requested to restart.

    Returns
    -------
    dict
        The parsed JSON response of the request's response as delivered by the
        diop.adapters.psytricks.ResTricksClient.
    """
    # TODO: refuse action if session information in DB is outdated!
    return psytricks.terminate_session(psy_session_uid=uid)


def disconnect_all_sessions(dry_run=False):
    """Send a "disconnect" request to all session in state "active".

    Parameters
    ----------
    dry_run : bool, optional
        If True no disconnect request will be sent, only a warning message will
        be logged instead. By default False.
    """
    sessions = Session.objects.filter(t_end=None)
    log.debug(f"🧾🖥🎬 Got {len(sessions)} session(s) from the DB.")
    for session in sessions:
        if session.cur_properties.state == "active":
            if dry_run:
                log.warning(f"DRY-RUN: session {session} would be disconnected...")
                continue
            _BROKER.disconnect_session(session.machine.fqdn)
            record = SiteLog(
                name="Session disconnected",
                session=session,
                user=session.user,
                machine=session.machine,
                deliverygroup=session.machine.deliverygroup,
                parameters=f"{session.user}@{session.machine.hostname}",
            )
            record.save()
        else:
            log.warning(f"Nothing to do for session {session}")


@fire_and_dont_forget
def disconnect_sessions_of(user):
    """Disconnect all sessions of the given username.

    Parameters
    ----------
    user : str
        The username for which sessions should be disconnected.
    """
    sessions = Session.objects.filter(t_end=None, user=user)
    log.debug(f"🧾🖥🎬 Got {len(sessions)} session(s) of [{user}] from the DB.")
    for session in sessions:
        _BROKER.disconnect_session(session.machine.fqdn)
        record = SiteLog(
            name="Session disconnected",
            session=session,
            user=session.user,
            machine=session.machine,
            deliverygroup=session.machine.deliverygroup,
            parameters=f"{session.user}@{session.machine.hostname}",
        )
        record.save()


@fire_and_dont_forget
def disconnect_session(uid):
    """Send a "disconnect" request to the session with the given UID.

    Parameters
    ----------
    uid : int
        The session UID that should be disconnected.
    """
    session = Session.objects.get(uid=uid)
    log.debug(f"🖥🎬📢 <disconnect> ⏩ [{uid}]@[{session.machine.fqdn}]")
    _BROKER.disconnect_session(session.machine.fqdn)
    record = SiteLog(
        name="Session disconnected",
        session=session,
        user=session.user,
        machine=session.machine,
        deliverygroup=session.machine.deliverygroup,
        parameters=f"{session.user}@{session.machine.hostname}",
    )
    record.save()


@fire_and_dont_forget
def send_popup_message(uid, message, title, style):
    """Send a pop-up message to the session with the given UID.

    Parameters
    ----------
    uid : int
        The UID of the session that should receive and show the pop-up message.
    message : str
        The message to show.
    title : str
        The title of the message
    style : str
        The message style. Refer to the [pyppms][pyppms] docs for details.
    """
    session = Session.objects.get(uid=uid)
    machine = session.machine.fqdn
    user = session.user
    notification = Notification(
        user=user,
        session=session,
        method="P",
        cc_adm=False,
        subject=title,
        body=message,
        creator="Interactive (web)",
    )
    notification.save()
    log.debug(f"🖥🎬📢 <send_message> ⏩ [{uid}]@[{machine}]")
    _BROKER.send_message(machine, message, title, style)
    notification.t_sent = make_aware(datetime.now())
    notification.save()


@status_update_task
def pull_ppms_users(force_refresh=False):
    """Update DIOP users table with PPMS details."""
    log.success("Requesting update of PPMS users...")
    ppms = BookingSystem()
    return ppms.pull_users(force_refresh)


@status_update_task
def pull_ppms_bookings(dt=None, days=1):
    """Update DIOP reservations from PPMS bookings.

    Parameters
    ----------
    dt : date, optional
        The date to pull PPMS bookings for, by default None which will result in
        today's bookings being fetched.
    days : int, optional
        The number of follow-up days to pull from PPMS (plus one), by default 1
        which will result in only the date given as `dt` being fetched.

    Returns
    -------
    int, int
        A tuple with counts of success / failed booking updates.
    """
    # don't put `date.today()` as the default parameter value, this will become
    # static when being run through django-q2:
    if not dt:
        dt = date.today()
    okay = fail = 0
    ppms = BookingSystem()
    for delta in range(days):
        pull_date = dt + timedelta(days=delta)
        d_okay, d_fail = ppms.pull_bookings(dt=pull_date)
        okay += d_okay
        fail += d_fail

    return okay, fail


def process_maintenance_requests():
    """Check the 'req_maintenance' Machine flag and set maintenance accordingly.

    This function is intended to be run regularly by a scheduled tasks, it will
    check the "maintenance requested" flag for each machine and will assemble
    a list of machines that should be placed in maintenance mode (skipping
    those machines that are having an active session as users won't be able to
    reconnect to them once a machine is in maintenance) and a list of machines
    that should have maintenance disabled.

    Eventually `set_maintenance()` is called with those two lists to actually
    request the maintenance state adjustments.
    """
    # enabling maintenance this way is only done if there is no session on it:
    enable_on = []
    want_maintenance = Machine.objects.filter(req_maintenance=True)
    for machine in want_maintenance:
        try:
            session = Session.objects.get(machine=machine, t_end=None)
            log.debug(
                f"Machine {machine} is having a session ({session}), not placing "
                "in maintenance mode!"
            )
        except Session.DoesNotExist:
            enable_on.append(machine)

    # disabling maintenance can be done independently of a session:
    disable_on = Machine.objects.filter(req_maintenance=False)

    psytricks.set_maintenance(disable_on=disable_on, enable_on=enable_on)


def check_session_constraints(max_parallel=1):
    """Call all session constraint checking functions.

    Parameters
    ----------
    max_parallel : int, optional
        The maximum number of parallel sessions allowed, by default 1.
    """
    check_session_max_parallel(max_parallel)


def check_session_disconnect_time():  # WIP
    """Check disconnected sessions exceeding the (per-user) time limit.

    Using `diop.models.session.Session.exceeds_disconnected_limit`
    """
    sessions = Session.objects.filter(t_end=None)
    for session in sessions:
        if session.exceeds_disconnected_limit:
            log.warning(f"Session [{session}] is above the disconnection limit!")
            send_support_email("ops-disconnect-limit", {"session": session})


def check_session_max_parallel(limit=1):
    """Check sessions for violations of the max-parallel-sessions constraint.

    Query the DB for running sessions (i.e. that do not have an end time) that
    belong to any of the delivery groups having the `unique_session` attribute
    set and count the number of sessions for each user.

    If a user is having more sessions than allowed, assemble some details and
    send a pop-up notification to *each* of those sessions.

    By running this with the frequent housekeeping tasks it is ensured that a
    user gets almost instant feedback on the multiple sessions after starting
    a new one.

    Parameters
    ----------
    limit : int, optional
        The maximum count of parallel sessions in the DG's having the
        `unique_session` attribute set allowed for each user, by default 1.
    """
    unique_sess_dgs = DeliveryGroup.objects.filter(unique_session=True)
    violating = (
        Session.objects.filter(t_end=None, machine__deliverygroup__in=unique_sess_dgs)
        .values("user")
        .annotate(session_count=Count("user"))
        .filter(session_count__gt=limit)
    )
    if not violating:
        log.debug("No session violating the max-parallel-sessions constraint.")
        return

    dg_names = ", ".join(unique_sess_dgs.values_list("dg_name", flat=True))

    for multi in violating:
        username = multi["user"]
        count = multi["session_count"]
        sessions = Session.objects.filter(user=username, t_end=None).order_by("uid")
        uid_machine = "- " + "\n- ".join([f"({x.uid}) {x.machine}" for x in sessions])
        log.debug(f"User [{username}] is having {count} parallel sessions!")
        context = {
            "user": User.objects.get(pk=username),
            "uid_machine": uid_machine,
            "unique_sesssion_group_names": dg_names,
        }
        for session in sessions:
            Notification.new_popup(
                psytricks.send_popup, session.uid, "multi-sessions", context
            )


def check_reservations():
    """Check reservations for all DG's requiring booking.

    For each Delivery Group that requires booking the corresponding reservations
    will be checked and the status of user access permissions will be updated in
    the Diop DB.

    See Also
    --------
    check_reservations_for_dg
    """
    booking_groups = DeliveryGroup.objects.filter(booking_required=True)
    for dg in booking_groups:
        check_reservations_for_dg(dg)
        psytricks.update_diop_dg_access(dg_name=dg.dg_name)


def check_reservations_for_dg(dg: DeliveryGroup):  # WIP
    """Check reservations, enable DG access, send warnings, terminate sessions.

    Check all reservations and perform the following tasks:

    -  Ensure each user with a valid reservation is having access to the
       delivery group (DG) of the same name.
    - If a reservation is about to expire (< 10 minutes), send a notification
      (pop-up and email) to the user.
    - If a user doesn't have a valid reservation any more:
      - terminate their session in the DG (warn only in dry-run mode)
      - remove access to the DG
    - Send a notification to users having a running reservation >5m but no
      session. This will also serve as a reminder to cancel remaining booked
      time after finishing a session.
    - Re-send the reminder after 30m, Cc admins.

    Parameters
    ----------
    dg : DeliveryGroup
        The DG to process.
    """
    log.debug(f"Checking reservations for group [{dg.dg_name}]...")
    users_with_res = Reservation.running_usernames(dg.dg_name)

    log.debug(f"Adjusting access permissions for DG [{dg.dg_name}]...")
    psytricks.set_dg_access(
        dg_name=dg.dg_name,
        usernames=users_with_res,
        keep_common=settings.DIOP_GRACEFUL,
    )

    check_stale_reservations(dg=dg, users=users_with_res)

    # get active sessions of the current DG:
    sessions = Session.objects.filter(machine__deliverygroup=dg, t_end=None)
    for session in sessions:
        check_reservation_ending_soon(session)

        if not session.user.accessgroup_member:
            log.info(
                f"User [{session.user.username}] is not a member of the accessgroup, "
                "not checking session for a valid reservation (might be an admin)."
            )
            continue

        if session.user.username in users_with_res:
            continue

        if in_dry_run_mode("terminating session without reservation"):
            sent = Notification.new_email(
                user=session.user,
                template="session-without-reservation",
                context={
                    "dg_name": dg.dg_name,
                    "session": session,
                },
                support="",
                session=session,
            )
            if sent:
                log.warning(
                    "Sent notification email on session without valid reservation: "
                    f"{session} [{dg}]"
                )
            continue

        log.warning(f"Terminating session {session}, no valid reservation!")
        psytricks.terminate_session(v_session=session)
        record = SiteLog(
            name="Session terminated (reservation expired)",
            session=session,
            user=session.user,
            machine=session.machine,
            deliverygroup=session.machine.deliverygroup,
        )
        record.save()
        sent = Notification.new_email(
            user=session.user,
            template="session-terminated",
            context={
                "dg_name": dg.dg_name,
                "session": session,
            },
            support="",
            session=session,
        )
        if sent:
            log.warning(
                "Sent notification email on terminated 🎫🛑🎬 session: "
                f"{session} [{dg}]"
            )


def check_stale_reservations(dg: DeliveryGroup, users: QuerySet):
    """Check for reservations that do not have a session.

    Identify reservations of the given users in the specified DG that do not
    have an active session in that DG and process them.

    Parameters
    ----------
    dg : DeliveryGroup
        _description_
    users : django.db.models.query.QuerySet
        A queryset containing users having a currently running reservation.
    """
    users_with_session = Session.objects.filter(
        machine__deliverygroup=dg, t_end=None
    ).values_list("user__username", flat=True)

    res_but_no_session = users.difference(users_with_session)
    log.debug(f"Reservations without corresponding session: {len(res_but_no_session)}")
    stale = Reservation.running(dg).filter(user__username__in=res_but_no_session)
    for reservation in stale:
        process_stale_reservation(reservation)


def process_stale_reservation(
    reservation: Reservation, soft: int = 5, hard: int = 30, remaining: int = 15
):
    """Process 'stale' reservations, i.e. NOT having a corresponding session.

    If the reservation start is more than the given "soft" limit minutes ago,
    send an email notification to the user, if it's more than the "hard" limit
    minutes ago, re-send the notification with Cc'ing admins.

    Parameters
    ----------
    reservation : Reservation
        The reservation object to check.
    soft : int, optional
        The "soft" limit in minutes, after which a notification is sent to the
        user, by default 5.
    hard : int, optional
        The "hard" limit in minutes, after which the notification will be
        re-sent to the user AND the admins, by default 30.
    remaining : int, optional
        The "remaining" limit in minutes: if a stale reservation has less than
        the given number of minutes (default=15) left, nothing will be done.
        This is intended to prevent sending pointless notifications to users for
        reservations that are about to expire anyway.

    Returns
    -------
    bool
        True in case a notification was sent, False otherwise.
    """
    # calculate limit values in seconds:
    lim_hard = hard * 60
    lim_soft = soft * 60
    lim_remaining = remaining * 60

    if reservation.till_end < lim_remaining:
        log.debug(
            f"Reservation {reservation} has less than {remaining}m left, ignoring."
        )
        return False

    since_start = reservation.since_start
    if since_start < lim_soft:
        log.debug(
            f"Reservation {reservation} only started {since_start}s ago, ignoring."
        )
        return False

    dg = reservation.deliverygroup
    user = reservation.user

    # default notification properties (will be overridden below if necessary):
    template = "stale-reservation"
    support = ""  # do not Cc admins
    context = {
        "reservation": reservation,
        "dg_name": dg.dg_name,
        "delta": soft,
    }

    # reservation start delta is above the hard limit, Cc admins:
    if since_start > lim_hard:
        context["delta"] = hard
        support = "cc"

    # check if there has been an actual session during the lifetime of this
    # reservation and send a reminder to cancel the leftover time:
    last_completed = Session.last_completed(dg=dg, user=user)
    log.trace(f"last_completed.t_end: {last_completed.t_end}")
    log.trace(f"reservation.t_start: {reservation.t_start}")
    if last_completed.t_end > reservation.t_start:
        support = ""  # do not Cc admins
        context["delta"] = remaining
        template = "remaining-reservation-time-please-cancel"
        context["last_completed"] = last_completed
        ended = (now() - last_completed.t_end).total_seconds()
        log.trace(f"Session related to res {reservation} ended {ended}s ago.")
        if ended > lim_hard:
            log.trace(f"Session end more than {hard}m ago, will Cc admins...")
            support = "cc"

    sent = Notification.new_email(
        user=reservation.user,
        template=template,
        context=context,
        support=support,
    )
    return sent


def check_reservation_ending_soon(session: Session, threshold: int = 10):
    """Check if a session's reservation is about to expire and send a warning.

    The given session will be checked for its corresponding reservation. In case
    the remaining runtime of the reservation is below the threshold a warning
    will be sent to the session owner (via pop-up and email).

    Parameters
    ----------
    session : Session
        The session to check the reservation for.
    threshold : int
        The remaining reservation time threshold in minutes, default=10.

    Returns
    -------
    bool
        True in case the remaining reservation time is below the given
        threshold, False otherwise.
    """
    dg = session.machine.deliverygroup
    reservation = Reservation.ending_in_less_than(
        minutes=threshold, dg_name=dg.dg_name, username=session.user.username
    ).first()  # only the first is required here (even if there were multiple)

    if not reservation:
        return False

    context = {
        "reservation": reservation,
        "session": session,
        "dg_name": reservation.deliverygroup.dg_name,
    }
    Notification.new_popup(
        psytricks.send_popup, session.uid, "reservation-ending-soon", context
    )
    Notification.new_email(session.user, "reservation-ending-soon", context)

    return True


@fire_and_dont_forget
def send_session_reminders():
    """Send a reminder email for each disconnected session."""
    sessions = Session.objects.filter(t_end=None)
    for session in sessions:
        if session.cur_state != "disconnected":
            continue

        log.debug(f"Sending session reminder for {session}...")
        context = {
            "session": session,
        }
        Notification.new_email(
            user=session.user,
            template="session-reminder",
            context=context,
            session=session,
        )


@status_update_task
def update_users_from_ad():
    """Update all users from AD that are members of the access-group.

    In step one, it is made sure for each user exists a corresponding Diop user
    record and the values of the attributes `fullname`, `department` and `email`
    is set according to the information from AD. In addition, their
    `accessgroup_member` attribute is set to `True`.

    In the second step the Diop DB is queried for entries whose usernames are
    NOT in the list of users of the AD access-group to make sure their
    `accessgroup_member` attribute is set to `False`.
    """
    ad_conn = ActiveDirectory()
    # fetch users from AD, this will also fill the `ad_conn.user_details` dict:
    ad_group_members = ad_conn.group_members_usernames

    # part 1: ensure user record exists and has consistent data:
    for ad_username in ad_group_members:
        ad_user = ad_conn.user_details[ad_username]
        User.is_up_to_date_from_ad(user_details=ad_user, accessgroup_member=True)

    # part 2: ensure all other accounts have 'accessgroup_member' set to False:
    non_members = User.objects.exclude(username__in=ad_group_members)
    for non_member in non_members:
        if non_member.accessgroup_member:
            log.debug(f"Accessgroup membership [{non_member.username}] -> False")
            non_member.accessgroup_member = False
            non_member.save()


@status_update_task
def housekeeping_every_minute():
    """Regular housekeeping tasks that should be done every minute.

    Run various checks on the status of the DIOP DB:
    - [ ] age of sessions and machines status
    - [ ] age of last PPMS sync
    - [ ] age of last LDAP sync
    - [ ] terminate long-disconnected sessions
    - [x] put a machine in maintenance mode if requested and no session
    - [x] double sessions logic (warn, then terminate after 1h?)
    - [x] enable / disable access to DG's depending on reservation status
    - [x] terminate sessions whose reservation expired
    - [ ] users whose AD account was disabled
    - ...
    """
    process_maintenance_requests()
    check_session_constraints()
    check_reservations()


@status_update_task
def housekeeping_start_of_day():
    """Housekeeping tasks to be run at the start of a day, e.g. at 09:00.

    - [ ] platform summary: sessions, storage, deltas, users, ...
    - [x] disconnection time reminder (-> once a day, not once a minute!)
    - [ ] users without a PPMS account (required for billing)
    """
    send_session_reminders()


@fire_and_dont_forget
def housekeeping_early_morning():
    """Housekeeping tasks to be run at the very early morning, e.g. at 04:00.

    The idea of these tasks is to run them at a time with the lowest probability
    of any user actively working.

    - [x] disconnect all active sessions
    """
    dry_run = in_dry_run_mode("disconnecting active sessions")
    disconnect_all_sessions(dry_run=dry_run)



# Generated by Django 4.1.7 on 2023-04-17 12:29

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("diop", "0005_machine_deliverygroup"),
    ]

    operations = [
        migrations.AlterField(
            model_name="deliverygroup",
            name="booking_required",
            field=models.BooleanField(
                default=False,
                help_text="Sessions in this DG require a valid booking.",
                verbose_name="Booking Required",
            ),
        ),
        migrations.AlterField(
            model_name="deliverygroup",
            name="dg_name",
            field=models.CharField(
                help_text="Name of the Delivery Group (DG).",
                max_length=32,
                primary_key=True,
                serialize=False,
                verbose_name="Group Name",
            ),
        ),
        migrations.AlterField(
            model_name="deliverygroup",
            name="host_prefix",
            field=models.CharField(
                help_text="Hostnames starting with this prefix will be considered members of the DG.",
                max_length=32,
                verbose_name="Hostname Prefix",
            ),
        ),
        migrations.AlterField(
            model_name="machine",
            name="deliverygroup",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="diop.deliverygroup",
                verbose_name="Delivery Group",
            ),
        ),
        migrations.AlterField(
            model_name="machine",
            name="fqdn",
            field=models.CharField(
                help_text="Fully Qualified Domain Name",
                max_length=64,
                primary_key=True,
                serialize=False,
                verbose_name="FQDN",
            ),
        ),
    ]
